@@ -21,13 +21,12 @@
 
 #include "proto.h"
 
+#include <errno.h>
 #include <stdint.h>
-#include <stdio.h>
 #include <string.h>
 #include <unistd.h>
-#include <errno.h>
 
-#ifndef DISABLE_BROWSER
+#ifdef ENABLE_BROWSER
 
 static char **filelist = NULL;
 	/* The list of files to display in the file browser. */
@@ -172,25 +171,27 @@ char *do_browser(char *path)
 	    say_there_is_no_help();
 #endif
 	} else if (func == do_search) {
-	    /* Search for a filename. */
 	    do_filesearch();
 	} else if (func == do_research) {
-	    /* Search for another filename. */
-	    do_fileresearch();
+	    do_fileresearch(TRUE);
+#ifndef NANO_TINY
+	} else if (func == do_findprevious) {
+	    do_fileresearch(FALSE);
+	} else if (func == do_findnext) {
+	    do_fileresearch(TRUE);
+#endif
 	} else if (func == do_left) {
 	    if (selected > 0)
 		selected--;
 	} else if (func == do_right) {
 	    if (selected < filelist_len - 1)
 		selected++;
-#ifndef NANO_TINY
 	} else if (func == do_prev_word_void) {
 	    selected -= (selected % width);
 	} else if (func == do_next_word_void) {
 	    selected += width - 1 - (selected % width);
 	    if (selected >= filelist_len)
 		selected = filelist_len - 1;
-#endif
 	} else if (func == do_up_void) {
 	    if (selected >= width)
 		selected -= width;
@@ -251,11 +252,10 @@ char *do_browser(char *path)
 	    }
 
 #ifndef DISABLE_OPERATINGDIR
-	    if (check_operating_dir(path, FALSE)) {
+	    if (outside_of_confinement(path, FALSE)) {
 		/* TRANSLATORS: This refers to the confining effect of the
 		 * option --operatingdir, not of --restricted. */
-		statusline(ALERT, _("Can't go outside of %s"),
-				full_operating_dir);
+		statusline(ALERT, _("Can't go outside of %s"), operating_dir);
 		path = mallocstrcpy(path, present_path);
 		continue;
 	    }
@@ -285,9 +285,8 @@ char *do_browser(char *path)
 	    /* Note: The selected file can be outside the operating
 	     * directory if it's ".." or if it's a symlink to a
 	     * directory outside the operating directory. */
-	    if (check_operating_dir(filelist[selected], FALSE)) {
-		statusline(ALERT, _("Can't go outside of %s"),
-				full_operating_dir);
+	    if (outside_of_confinement(filelist[selected], FALSE)) {
+		statusline(ALERT, _("Can't go outside of %s"), operating_dir);
 		continue;
 	    }
 #endif
@@ -384,7 +383,7 @@ char *do_browse_from(const char *inpath)
 #ifndef DISABLE_OPERATINGDIR
     /* If the resulting path isn't in the operating directory, use
      * the operating directory instead. */
-    if (check_operating_dir(path, FALSE))
+    if (outside_of_confinement(path, FALSE))
 	path = mallocstrcpy(path, operating_dir);
 #endif
 
@@ -493,6 +492,9 @@ functionptrtype parse_browser_input(int *kbinput)
 	    case '/':
 		return do_search;
 	    case 'N':
+#ifndef NANO_TINY
+		return do_findprevious;
+#endif
 	    case 'n':
 		return do_research;
 	}
@@ -541,7 +543,7 @@ void browser_refresh(void)
 	/* If this is the selected item, start its highlighting, and
 	 * remember its location to be able to place the cursor on it. */
 	if (i == selected) {
-	    wattron(edit, hilite_attribute);
+	    wattron(edit, interface_color_pair[SELECTED_TEXT]);
 	    the_row = row;
 	    the_column = col;
 	}
@@ -613,7 +615,7 @@ void browser_refresh(void)
 
 	/* If this is the selected item, finish its highlighting. */
 	if (i == selected)
-	    wattroff(edit, hilite_attribute);
+	    wattroff(edit, interface_color_pair[SELECTED_TEXT]);
 
 	free(info);
 
@@ -702,8 +704,9 @@ int filesearch_init(void)
     return input;
 }
 
-/* Look for the given needle in the list of files. */
-void findnextfile(const char *needle)
+/* Look for the given needle in the list of files.  If forwards is TRUE,
+ * search forward in the list; otherwise, search backward. */
+void findfile(const char *needle, bool forwards)
 {
     size_t looking_at = selected;
 	/* The location in the file list of the filename we're looking at. */
@@ -723,12 +726,16 @@ void findnextfile(const char *needle)
     /* Step through each filename in the list until a match is found or
      * we've come back to the point where we started. */
     while (TRUE) {
-	/* Move to the next filename in the list, or back to the first. */
-	if (looking_at < filelist_len - 1)
-	    looking_at++;
-	else {
-	    looking_at = 0;
-	    statusbar(_("Search Wrapped"));
+	if (forwards) {
+	    if (looking_at++ == filelist_len - 1) {
+		looking_at = 0;
+		statusbar(_("Search Wrapped"));
+	    }
+	} else {
+	    if (looking_at-- == 0) {
+		looking_at = filelist_len - 1;
+		statusbar(_("Search Wrapped"));
+	    }
 	}
 
 	/* Get the bare filename, without the path. */
@@ -775,16 +782,17 @@ void do_filesearch(void)
 	update_history(&search_history, answer);
 #endif
 
-    findnextfile(answer);
+    findfile(answer, TRUE);
 }
 
-/* Search again for the last given filename, without prompting. */
-void do_fileresearch(void)
+/* Search again without prompting for the last given search string,
+ * either forwards or backwards. */
+void do_fileresearch(bool forwards)
 {
     if (*last_search == '\0')
 	statusbar(_("No current search pattern"));
     else
-	findnextfile(last_search);
+	findfile(last_search, forwards);
 }
 
 /* Select the first file in the list. */
@@ -812,4 +820,4 @@ char *strip_last_component(const char *path)
     return copy;
 }
 
-#endif /* !DISABLE_BROWSER */
+#endif /* ENABLE_BROWSER */
