@@ -21,8 +21,6 @@
 
 #include "proto.h"
 
-#include <stdio.h>
-#include <stdarg.h>
 #include <string.h>
 
 static char *prompt = NULL;
@@ -57,7 +55,7 @@ int do_statusbar_input(bool *ran_func, bool *finished)
 	return KEY_WINCH;
 #endif
 
-#ifndef DISABLE_MOUSE
+#ifdef ENABLE_MOUSE
     /* If we got a mouse click and it was on a shortcut, read in the
      * shortcut character. */
     if (input == KEY_MOUSE) {
@@ -95,7 +93,7 @@ int do_statusbar_input(bool *ran_func, bool *finished)
 	    kbinput = (int *)nrealloc(kbinput, kbinput_len * sizeof(int));
 	    kbinput[kbinput_len - 1] = input;
 	}
-     }
+    }
 
     /* If we got a shortcut, or if there aren't any other keystrokes waiting
      * after the one we read in, we need to insert all the characters in the
@@ -124,9 +122,9 @@ int do_statusbar_input(bool *ran_func, bool *finished)
 	else if (s->scfunc == do_next_word_void)
 	    do_statusbar_next_word();
 #endif
-	else if (s->scfunc == do_home_void)
+	else if (s->scfunc == do_home)
 	    do_statusbar_home();
-	else if (s->scfunc == do_end_void)
+	else if (s->scfunc == do_end)
 	    do_statusbar_end();
 	/* When in restricted mode at the "Write File" prompt and the
 	 * filename isn't blank, disallow any input and deletion. */
@@ -137,15 +135,18 @@ int do_statusbar_input(bool *ran_func, bool *finished)
 				s->scfunc == do_delete ||
 				s->scfunc == do_backspace))
 	    ;
-	else if (s->scfunc == do_verbatim_input) {
+	else if (s->scfunc == do_verbatim_input)
 	    do_statusbar_verbatim_input();
-	} else if (s->scfunc == do_cut_text_void)
+	else if (s->scfunc == do_cut_text_void)
 	    do_statusbar_cut_text();
 	else if (s->scfunc == do_delete)
 	    do_statusbar_delete();
 	else if (s->scfunc == do_backspace)
 	    do_statusbar_backspace();
-	else {
+	else if (s->scfunc == do_uncut_text) {
+	    if (cutbuffer != NULL)
+		do_statusbar_uncut_text();
+	} else {
 	    /* Handle any other shortcut in the current menu, setting
 	     * ran_func to TRUE if we try to run their associated functions,
 	     * and setting finished to TRUE to indicatethat we're done after
@@ -164,7 +165,7 @@ int do_statusbar_input(bool *ran_func, bool *finished)
     return input;
 }
 
-#ifndef DISABLE_MOUSE
+#ifdef ENABLE_MOUSE
 /* Handle a mouse click on the statusbar prompt or the shortcut list. */
 int do_statusbar_mouse(void)
 {
@@ -291,7 +292,7 @@ void do_statusbar_delete(void)
 /* Zap some or all text from the answer. */
 void do_statusbar_cut_text(void)
 {
-    if (!ISSET(CUT_TO_END))
+    if (!ISSET(CUT_FROM_CURSOR))
 	statusbar_x = 0;
 
     answer[statusbar_x] = '\0';
@@ -363,6 +364,23 @@ size_t statusbar_xplustabs(void)
     return strnlenpt(answer, statusbar_x);
 }
 
+/* Paste the first line of the cutbuffer into the current answer. */
+void do_statusbar_uncut_text(void)
+{
+    size_t pastelen = strlen(cutbuffer->data);
+    char *fusion = charalloc(strlen(answer) + pastelen + 1);
+
+    /* Concatenate: the current answer before the cursor, the first line
+     * of the cutbuffer, plus the rest of the current answer. */
+    strncpy(fusion, answer, statusbar_x);
+    strncpy(fusion + statusbar_x, cutbuffer->data, pastelen);
+    strcpy(fusion + statusbar_x + pastelen, answer + statusbar_x);
+
+    free(answer);
+    answer = fusion;
+    statusbar_x += pastelen;
+}
+
 /* Return the column number of the first character of the answer that is
  * displayed in the statusbar when the cursor is at the given column,
  * with the available room for the answer starting at base.  Note that
@@ -383,35 +401,18 @@ void reinit_statusbar_x(void)
     statusbar_x = HIGHEST_POSITIVE;
 }
 
-/* Put the cursor in the answer at statusbar_x. */
-void reset_statusbar_cursor(void)
-{
-    size_t start_col = strlenpt(prompt) + 2;
-    size_t xpt = statusbar_xplustabs();
-
-    /* Work around a cursor-misplacement bug in VTEs. */
-    wmove(bottomwin, 0, 0);
-    wnoutrefresh(bottomwin);
-    doupdate();
-
-    wmove(bottomwin, 0, start_col + xpt -
-			get_statusbar_page_start(start_col, start_col + xpt));
-
-    wnoutrefresh(bottomwin);
-}
-
-/* Repaint the statusbar. */
+/* Redraw the promptbar and place the cursor at the right spot. */
 void update_the_statusbar(void)
 {
-    size_t base, the_page, end_page;
+    size_t base = strlenpt(prompt) + 2;
+    size_t the_page, end_page, column;
     char *expanded;
 
-    base = strlenpt(prompt) + 2;
     the_page = get_statusbar_page_start(base, base + strnlenpt(answer, statusbar_x));
     end_page = get_statusbar_page_start(base, base + strlenpt(answer) - 1);
 
+    /* Color the promptbar over its full width. */
     wattron(bottomwin, interface_color_pair[TITLE_BAR]);
-
     blank_statusbar();
 
     mvwaddstr(bottomwin, 0, 0, prompt);
@@ -426,7 +427,14 @@ void update_the_statusbar(void)
 
     wattroff(bottomwin, interface_color_pair[TITLE_BAR]);
 
-    reset_statusbar_cursor();
+    /* Work around a cursor-misplacement bug in VTEs. */
+    wmove(bottomwin, 0, 0);
+    wrefresh(bottomwin);
+
+    /* Place the cursor at statusbar_x in the answer. */
+    column = base + statusbar_xplustabs();
+    wmove(bottomwin, 0, column - get_statusbar_page_start(base, column));
+    wnoutrefresh(bottomwin);
 }
 
 /* Get a string of input at the statusbar prompt. */
@@ -440,7 +448,7 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
     int kbinput = ERR;
     bool ran_func, finished;
     functionptrtype func;
-#ifndef DISABLE_TABCOMP
+#ifdef ENABLE_TABCOMP
     bool tabbed = FALSE;
 	/* Whether we've pressed Tab. */
 #endif
@@ -450,7 +458,7 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
     char *magichistory = NULL;
 	/* The temporary string typed at the bottom of the history, if
 	 * any. */
-#ifndef DISABLE_TABCOMP
+#ifdef ENABLE_TABCOMP
     int last_kbinput = ERR;
 	/* The key we pressed before the current key. */
     size_t complete_len = 0;
@@ -468,10 +476,6 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
 
     update_the_statusbar();
 
-    /* Refresh edit window and statusbar before getting input. */
-    wnoutrefresh(edit);
-    wnoutrefresh(bottomwin);
-
     while (TRUE) {
 	/* Ensure the cursor is shown when waiting for input. */
 	curs_set(1);
@@ -495,7 +499,7 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
 	if (func == do_cancel || func == do_enter)
 	    break;
 
-#ifndef DISABLE_TABCOMP
+#ifdef ENABLE_TABCOMP
 	if (func != do_tab)
 	    tabbed = FALSE;
 
@@ -516,7 +520,7 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
 		answer = input_tab(answer, allow_files, &statusbar_x,
 					&tabbed, refresh_func, listed);
 	} else
-#endif /* !DISABLE_TABCOMP */
+#endif /* ENABLE_TABCOMP */
 #ifndef DISABLE_HISTORIES
 	if (func == get_history_older_void) {
 	    if (history_list != NULL) {
@@ -580,7 +584,7 @@ functionptrtype acquire_an_answer(int *actual, bool allow_tabs,
 
 	update_the_statusbar();
 
-#if !defined(DISABLE_HISTORIES) && !defined(DISABLE_TABCOMP)
+#if !defined(DISABLE_HISTORIES) && defined(ENABLE_TABCOMP)
 	last_kbinput = kbinput;
 #endif
     }
@@ -620,8 +624,9 @@ int do_prompt(bool allow_tabs, bool allow_files,
     int retval;
     functionptrtype func = NULL;
     bool listed = FALSE;
-    /* Save a possible current statusbar x position. */
+    /* Save a possible current statusbar x position and prompt. */
     size_t was_statusbar_x = statusbar_x;
+    char *saved_prompt = prompt;
 
     bottombars(menu);
 
@@ -644,7 +649,7 @@ int do_prompt(bool allow_tabs, bool allow_files,
 			refresh_func);
 
     free(prompt);
-    prompt = NULL;
+    prompt = saved_prompt;
 
 #ifndef NANO_TINY
     if (retval == KEY_WINCH)
@@ -670,7 +675,7 @@ int do_prompt(bool allow_tabs, bool allow_files,
     fprintf(stderr, "answer = \"%s\"\n", answer);
 #endif
 
-#ifndef DISABLE_TABCOMP
+#ifdef ENABLE_TABCOMP
     /* If we've done tab completion, there might still be a list of
      * filename matches on the edit window.  Clear them off. */
     if (listed)
@@ -704,7 +709,7 @@ int do_yesno_prompt(bool all, const char *msg)
 	functionptrtype func;
 
 	if (!ISSET(NO_HELP)) {
-	    char shortstr[3];
+	    char shortstr[MAXCHARLEN + 2];
 		/* Temporary string for (translated) " Y", " N" and " A". */
 
 	    if (COLS < 32)
@@ -751,7 +756,7 @@ int do_yesno_prompt(bool all, const char *msg)
 
 	if (func == do_cancel)
 	    response = -1;
-#ifndef DISABLE_MOUSE
+#ifdef ENABLE_MOUSE
 	else if (kbinput == KEY_MOUSE) {
 	    int mouse_x, mouse_y;
 	    /* We can click on the Yes/No/All shortcuts to select an answer. */
@@ -773,7 +778,7 @@ int do_yesno_prompt(bool all, const char *msg)
 		    response = -2;
 	    }
 	}
-#endif /* !DISABLE_MOUSE */
+#endif /* ENABLE_MOUSE */
 	else {
 	    /* Look for the kbinput in the Yes, No (and All) strings. */
 	    if (strchr(yesstr, kbinput) != NULL)
